/* -*- mode: c++; indent-tabs-mode: nil -*- */
/** @file SFTPClient.qpp defines the SFTPClient class */
/*
    SFTPClient.qpp

    libssh2 SFTP client integration into qore

    Copyright 2009 Wolfgang Ritzinger
    Copyright 2010 - 2021 Qore Technologies, s.r.o.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "SFTPClient.h"
#include "QC_SSH2Base.h"

static QoreHashNode* attr2hash(const LIBSSH2_SFTP_ATTRIBUTES& attr, ExceptionSink* xsink) {
    ReferenceHolder<QoreHashNode> ret(new QoreHashNode(hashdeclSsh2StatInfo, xsink), xsink);

    if (attr.flags & LIBSSH2_SFTP_ATTR_SIZE)
        ret->setKeyValue("size", attr.filesize, xsink);
    if (attr.flags & LIBSSH2_SFTP_ATTR_ACMODTIME) {
        ret->setKeyValue("atime", DateTimeNode::makeAbsolute(currentTZ(), (int64)attr.atime), xsink);
        ret->setKeyValue("mtime", DateTimeNode::makeAbsolute(currentTZ(), (int64)attr.mtime), xsink);
    }
    if (attr.flags & LIBSSH2_SFTP_ATTR_UIDGID) {
        ret->setKeyValue("uid", attr.uid, xsink);
        ret->setKeyValue("gid", attr.gid, xsink);
    }
    if (attr.flags & LIBSSH2_SFTP_ATTR_PERMISSIONS) {
        ret->setKeyValue("mode", attr.permissions, xsink);
        ret->setKeyValue("permissions", new QoreStringNode(mode2str(attr.permissions)), xsink);
    }

    return ret.release();
}
TypedHashDecl* init_hashdecl_SftpFileInfo(QoreNamespace& ns) {
    TypedHashDecl* hd = new TypedHashDecl("SftpFileInfo", "::Qore::SftpFileInfo");
    hd->addMember("atime", dateTypeInfo, QoreValue());
    hd->addMember("gid", bigIntTypeInfo, QoreValue());
    hd->addMember("mode", bigIntTypeInfo, QoreValue());
    hd->addMember("mtime", dateTypeInfo, QoreValue());
    hd->addMember("name", stringTypeInfo, QoreValue());
    hd->addMember("perm", stringTypeInfo, QoreValue());
    hd->addMember("size", bigIntTypeInfo, QoreValue());
    hd->addMember("type", stringTypeInfo, QoreValue());
    hd->addMember("uid", bigIntTypeInfo, QoreValue());
    ns.addSystemHashDecl(hd);
    return hd;
}

TypedHashDecl* init_hashdecl_SftpDirInfo(QoreNamespace& ns) {
    TypedHashDecl* hd = new TypedHashDecl("SftpDirInfo", "::Qore::SftpDirInfo");
    hd->addMember("directories", qore_get_complex_list_type(stringTypeInfo), QoreValue());
    hd->addMember("files", qore_get_complex_list_type(stringTypeInfo), QoreValue());
    hd->addMember("links", qore_get_complex_list_type(stringTypeInfo), QoreValue());
    hd->addMember("path", stringTypeInfo, QoreValue());
    ns.addSystemHashDecl(hd);
    return hd;
}

TypedHashDecl* init_hashdecl_SftpConnectionInfo(QoreNamespace& ns) {
    TypedHashDecl* hd = new TypedHashDecl("SftpConnectionInfo", "::Qore::SftpConnectionInfo");
    hd->addMember("authenticated", stringOrNothingTypeInfo, QoreValue());
    hd->addMember("connected", boolTypeInfo, QoreValue());
    hd->addMember("fingerprint", stringOrNothingTypeInfo, QoreValue());
    hd->addMember("keyfile_priv", stringTypeInfo, QoreValue());
    hd->addMember("keyfile_pub", stringTypeInfo, QoreValue());
    hd->addMember("methods", qore_get_complex_hash_type(stringTypeInfo), QoreValue());
    hd->addMember("path", stringOrNothingTypeInfo, QoreValue());
    hd->addMember("ssh2host", stringTypeInfo, QoreValue());
    hd->addMember("ssh2port", bigIntTypeInfo, QoreValue());
    hd->addMember("ssh2user", stringTypeInfo, QoreValue());
    ns.addSystemHashDecl(hd);
    return hd;
}

/* Qore class Qore::SSH2::SFTPClient */

qore_classid_t CID_SFTPCLIENT;
QoreClass* QC_SFTPCLIENT;

// string SFTPClient::chdir(string path, timeout timeout = 60s){}
static QoreValue SFTPClient_chdir_VsVt(QoreObject* self, SFTPClient* myself, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* path = HARD_QORE_VALUE_STRING(args, 0);
    int64 timeout = HARD_QORE_VALUE_INT(args, 1);
# 626 "QC_SFTPClient.qpp"
    return myself->sftpChdir(path->c_str(), (int)timeout, xsink);
}

// nothing SFTPClient::chmod(string path, int mode, timeout timeout = 60s){}
static QoreValue SFTPClient_chmod_VsViVt(QoreObject* self, SFTPClient* myself, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* path = HARD_QORE_VALUE_STRING(args, 0);
    int64 mode = HARD_QORE_VALUE_INT(args, 1);
    int64 timeout = HARD_QORE_VALUE_INT(args, 2);
# 338 "QC_SFTPClient.qpp"
    // check if mode is in range
    if (mode != (mode & (int64)SFTP_UGOMASK)) {
        xsink->raiseException("SFTPCLIENT-PARAMETER-ERROR", "mode setting is only possible for user, group and other (no sticky bits)");
        return QoreValue();
    }

    myself->sftpChmod(path->c_str(), (int)mode, (int)timeout, xsink);
    return QoreValue();
}

// SFTPClient::constructor(string url, softint port = 0) {}
static void SFTPClient_constructor_Vsvi(QoreObject* self, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* url = HARD_QORE_VALUE_STRING(args, 0);
    int64 port = HARD_QORE_VALUE_INT(args, 1);
# 149 "QC_SFTPClient.qpp"
QoreURL qurl(url);

    if (!qurl.getHost()) {
        xsink->raiseException("SFTPCLIENT-PARAMETER-ERROR", "no hostname found in URL '%s'", url->c_str());
        return;
    }

    if (qurl.getProtocol() && strcasecmp("sftp", qurl.getProtocol()->c_str())) {
        xsink->raiseException("SFTPCLIENT-PARAMETER-ERROR", "URL given in the first argument to SFTPClient::constructor() specifies invalid protocol '%s' (expecting 'sftp')", qurl.getProtocol()->getBuffer());
        return;
    }

    // create private data object
    self->setPrivate(CID_SFTPCLIENT, new SFTPClient(qurl, (int)port));
}

// SFTPClient::copy() {}
static void SFTPClient_copy(QoreObject* self, QoreObject* old, SFTPClient* myself, ExceptionSink* xsink) {
# 169 "QC_SFTPClient.qpp"
xsink->raiseException("SFTPCLIENT-COPY-ERROR", "copying sftp connection objects is not allowed");
}

// int SFTPClient::get(string remote_path, Qore::OutputStream os, timeout timeout = 60s){}
static QoreValue SFTPClient_get_VsC12OutputStreamVt(QoreObject* self, SFTPClient* myself, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* remote_path = HARD_QORE_VALUE_STRING(args, 0);
    HARD_QORE_VALUE_OBJ_DATA(os, OutputStream, args, 1, CID_OUTPUTSTREAM, "SFTPClient::get()", "Qore::OutputStream", xsink);
    if (*xsink)
        return 0;
    int64 timeout = HARD_QORE_VALUE_INT(args, 2);
# 370 "QC_SFTPClient.qpp"
    SimpleRefHolder<OutputStream> osHolder(os);
    return myself->sftpGet(remote_path->c_str(), os, (int) timeout, xsink);
}

// binary SFTPClient::getFile(string path, timeout timeout = 60s){}
static QoreValue SFTPClient_getFile_VsVt(QoreObject* self, SFTPClient* myself, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* path = HARD_QORE_VALUE_STRING(args, 0);
    int64 timeout = HARD_QORE_VALUE_INT(args, 1);
# 396 "QC_SFTPClient.qpp"
    return myself->sftpGetFile(path->c_str(), (int)timeout, xsink);
}

// string SFTPClient::getTextFile(string path, timeout timeout = 60s, *string encoding){}
static QoreValue SFTPClient_getTextFile_VsVtNs(QoreObject* self, SFTPClient* myself, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* path = HARD_QORE_VALUE_STRING(args, 0);
    int64 timeout = HARD_QORE_VALUE_INT(args, 1);
    const QoreStringNode* encoding = get_param_value(args, 2).get<const QoreStringNode>();
# 422 "QC_SFTPClient.qpp"
    const QoreEncoding *qe = encoding ? QEM.findCreate(encoding) : QCS_DEFAULT;
    return myself->sftpGetTextFile(path->c_str(), (int)timeout, qe, xsink);
}

// hash<SftpConnectionInfo> SFTPClient::info(){}
static QoreValue SFTPClient_info(QoreObject* self, SFTPClient* myself, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 191 "QC_SFTPClient.qpp"
    return myself->sftpInfo(xsink);
}

// bool SFTPClient::isAlive(timeout timeout_ms = 60){}
static QoreValue SFTPClient_isAlive_Vt(QoreObject* self, SFTPClient* myself, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    int64 timeout_ms = HARD_QORE_VALUE_INT(args, 0);
# 644 "QC_SFTPClient.qpp"
    return myself->sftpIsAlive((int)timeout_ms);
}

// bool SFTPClient::isAliveEx(timeout timeout_ms = 60){}
static QoreValue SFTPClient_isAliveEx_Vt(QoreObject* self, SFTPClient* myself, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    int64 timeout_ms = HARD_QORE_VALUE_INT(args, 0);
# 665 "QC_SFTPClient.qpp"
    return myself->sftpIsAliveEx((int)timeout_ms, xsink);
}

// hash<SftpDirInfo> SFTPClient::list(*string path, timeout timeout = 60s){}
static QoreValue SFTPClient_list_NsVt(QoreObject* self, SFTPClient* myself, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* path = get_param_value(args, 0).get<const QoreStringNode>();
    int64 timeout = HARD_QORE_VALUE_INT(args, 1);
# 226 "QC_SFTPClient.qpp"
    return myself->sftpList(path ? path->c_str() : nullptr, (int)timeout, xsink);
}

// list<hash<SftpFileInfo>> SFTPClient::listFull(*string path, timeout timeout = 60s){}
static QoreValue SFTPClient_listFull_NsVt(QoreObject* self, SFTPClient* myself, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* path = get_param_value(args, 0).get<const QoreStringNode>();
    int64 timeout = HARD_QORE_VALUE_INT(args, 1);
# 258 "QC_SFTPClient.qpp"
    return myself->sftpListFull(path ? path->c_str() : nullptr, (int)timeout, xsink);
}

// nothing SFTPClient::mkdir(string path, int mode = 0755, timeout timeout = 60s){}
static QoreValue SFTPClient_mkdir_VsViVt(QoreObject* self, SFTPClient* myself, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* path = HARD_QORE_VALUE_STRING(args, 0);
    int64 mode = HARD_QORE_VALUE_INT(args, 1);
    int64 timeout = HARD_QORE_VALUE_INT(args, 2);
# 593 "QC_SFTPClient.qpp"
    myself->sftpMkdir(path->c_str(), (int)mode, (int)timeout, xsink);
    return QoreValue();
}

// *string SFTPClient::path(){}
static QoreValue SFTPClient_path(QoreObject* self, SFTPClient* myself, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 201 "QC_SFTPClient.qpp"
    return myself->sftpPath();
}

// int SFTPClient::put(Qore::InputStream is, string remote_path, timeout timeout = 60s, int mode = 0644){}
static QoreValue SFTPClient_put_C11InputStreamVsVtVi(QoreObject* self, SFTPClient* myself, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    HARD_QORE_VALUE_OBJ_DATA(is, InputStream, args, 0, CID_INPUTSTREAM, "SFTPClient::put()", "Qore::InputStream", xsink);
    if (*xsink)
        return 0;
    const QoreStringNode* remote_path = HARD_QORE_VALUE_STRING(args, 1);
    int64 timeout = HARD_QORE_VALUE_INT(args, 2);
    int64 mode = HARD_QORE_VALUE_INT(args, 3);
# 574 "QC_SFTPClient.qpp"
    SimpleRefHolder<InputStream> isHolder(is);
    return myself->sftpPut(is, remote_path->c_str(), (int) mode, (int) timeout, xsink);
}

// int SFTPClient::putFile(binary bin, string path, int mode = 0644, timeout timeout = 60s){}
static QoreValue SFTPClient_putFile_VyVsViVt(QoreObject* self, SFTPClient* myself, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const BinaryNode* bin = HARD_QORE_VALUE_BINARY(args, 0);
    const QoreStringNode* path = HARD_QORE_VALUE_STRING(args, 1);
    int64 mode = HARD_QORE_VALUE_INT(args, 2);
    int64 timeout = HARD_QORE_VALUE_INT(args, 3);
# 482 "QC_SFTPClient.qpp"
    // transfer the file
    return myself->sftpPutFile((const char *)bin->getPtr(), bin->size(), path->c_str(), (int)mode, (int)timeout, xsink);
}

// int SFTPClient::putFile(string data, string path, int mode = 0644, timeout timeout = 60s){}
static QoreValue SFTPClient_putFile_VsVsViVt(QoreObject* self, SFTPClient* myself, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* data = HARD_QORE_VALUE_STRING(args, 0);
    const QoreStringNode* path = HARD_QORE_VALUE_STRING(args, 1);
    int64 mode = HARD_QORE_VALUE_INT(args, 2);
    int64 timeout = HARD_QORE_VALUE_INT(args, 3);
# 512 "QC_SFTPClient.qpp"
    // transfer the file
    return myself->sftpPutFile(data->c_str(), data->strlen(), path->c_str(), (int)mode, (int)timeout, xsink);
}

// nothing SFTPClient::removeFile(string path, timeout timeout = 60s){}
static QoreValue SFTPClient_removeFile_VsVt(QoreObject* self, SFTPClient* myself, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* path = HARD_QORE_VALUE_STRING(args, 0);
    int64 timeout = HARD_QORE_VALUE_INT(args, 1);
# 302 "QC_SFTPClient.qpp"
    myself->sftpUnlink(path->c_str(), (int)timeout, xsink);
    return QoreValue();
}

// nothing SFTPClient::rename(string old_name, string new_name, timeout timeout = 60s){}
static QoreValue SFTPClient_rename_VsVsVt(QoreObject* self, SFTPClient* myself, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* old_name = HARD_QORE_VALUE_STRING(args, 0);
    const QoreStringNode* new_name = HARD_QORE_VALUE_STRING(args, 1);
    int64 timeout = HARD_QORE_VALUE_INT(args, 2);
# 320 "QC_SFTPClient.qpp"
    myself->sftpRename(old_name->c_str(), new_name->c_str(), (int)timeout, xsink);
    return QoreValue();
}

// int SFTPClient::retrieveFile(string remote_path, string local_path, timeout timeout = 60s, int mode = 0644){}
static QoreValue SFTPClient_retrieveFile_VsVsVtVi(QoreObject* self, SFTPClient* myself, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* remote_path = HARD_QORE_VALUE_STRING(args, 0);
    const QoreStringNode* local_path = HARD_QORE_VALUE_STRING(args, 1);
    int64 timeout = HARD_QORE_VALUE_INT(args, 2);
    int64 mode = HARD_QORE_VALUE_INT(args, 3);
# 453 "QC_SFTPClient.qpp"
    return myself->sftpRetrieveFile(remote_path->c_str(), local_path->c_str(), (int)timeout, mode, xsink);
}

// nothing SFTPClient::rmdir(string path, timeout timeout = 60s){}
static QoreValue SFTPClient_rmdir_VsVt(QoreObject* self, SFTPClient* myself, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* path = HARD_QORE_VALUE_STRING(args, 0);
    int64 timeout = HARD_QORE_VALUE_INT(args, 1);
# 610 "QC_SFTPClient.qpp"
    myself->sftpRmdir(path->c_str(), (int)timeout, xsink);
    return QoreValue();
}

// *hash<Ssh2StatInfo> SFTPClient::stat(string path, timeout timeout = 60s){}
static QoreValue SFTPClient_stat_VsVt(QoreObject* self, SFTPClient* myself, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* path = HARD_QORE_VALUE_STRING(args, 0);
    int64 timeout = HARD_QORE_VALUE_INT(args, 1);
# 284 "QC_SFTPClient.qpp"
    LIBSSH2_SFTP_ATTRIBUTES attr;
    int rc = myself->sftpGetAttributes(path->c_str(), &attr, (int)timeout, xsink);
    return rc < 0 ? QoreValue() : attr2hash(attr, xsink);
}

// int SFTPClient::transferFile(string local_path, string remote_path, timeout timeout = 60s, *int mode){}
static QoreValue SFTPClient_transferFile_VsVsVtNi(QoreObject* self, SFTPClient* myself, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* local_path = HARD_QORE_VALUE_STRING(args, 0);
    const QoreStringNode* remote_path = HARD_QORE_VALUE_STRING(args, 1);
    int64 timeout = HARD_QORE_VALUE_INT(args, 2);
    int64 mode = HARD_QORE_VALUE_INT(args, 3);
# 543 "QC_SFTPClient.qpp"
    // transfer the file
    return myself->sftpTransferFile(local_path->c_str(), remote_path->c_str(), (int)mode, (int)timeout, xsink);
}

DLLLOCAL void preinitSFTPClientClass() {
    QC_SFTPCLIENT = new QoreClass("SFTPClient", "::Qore::SSH2::SFTPClient", QDOM_NETWORK);
    CID_SFTPCLIENT = QC_SFTPCLIENT->getID();
    QC_SFTPCLIENT->setSystem();
}

DLLLOCAL QoreClass* initSFTPClientClass(QoreNamespace& ns) {
    if (!QC_SFTPCLIENT)
        preinitSFTPClientClass();

    // set parent class
    assert(QC_SSH2BASE);
    QC_SFTPCLIENT->addBuiltinVirtualBaseClass(QC_SSH2BASE);

    // string SFTPClient::chdir(string path, timeout timeout = 60s){}
    QC_SFTPCLIENT->addMethod("chdir", (q_method_n_t)SFTPClient_chdir_VsVt, Public, QCF_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 2, stringTypeInfo, QORE_PARAM_NO_ARG, "path", timeoutTypeInfo, QoreSimpleValue().assign(DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0)), "timeout");

    // nothing SFTPClient::chmod(string path, int mode, timeout timeout = 60s){}
    QC_SFTPCLIENT->addMethod("chmod", (q_method_n_t)SFTPClient_chmod_VsViVt, Public, QCF_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 3, stringTypeInfo, QORE_PARAM_NO_ARG, "path", bigIntTypeInfo, QORE_PARAM_NO_ARG, "mode", timeoutTypeInfo, QoreSimpleValue().assign(DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0)), "timeout");

    // SFTPClient::constructor(string url, softint port = 0) {}
    QC_SFTPCLIENT->addConstructor(SFTPClient_constructor_Vsvi, Public, QCF_NO_FLAGS, QDOM_DEFAULT, 2, stringTypeInfo, QORE_PARAM_NO_ARG, "url", softBigIntTypeInfo, QoreSimpleValue().assign((int64)0), "port");

    // SFTPClient::copy() {}
    QC_SFTPCLIENT->setCopy((q_copy_t)SFTPClient_copy);

    // int SFTPClient::get(string remote_path, Qore::OutputStream os, timeout timeout = 60s){}
    QC_SFTPCLIENT->addMethod("get", (q_method_n_t)SFTPClient_get_VsC12OutputStreamVt, Public, QCF_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 3, stringTypeInfo, QORE_PARAM_NO_ARG, "remote_path", QC_OUTPUTSTREAM->getTypeInfo(), QORE_PARAM_NO_ARG, "os", timeoutTypeInfo, QoreSimpleValue().assign(DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0)), "timeout");

    // binary SFTPClient::getFile(string path, timeout timeout = 60s){}
    QC_SFTPCLIENT->addMethod("getFile", (q_method_n_t)SFTPClient_getFile_VsVt, Public, QCF_NO_FLAGS, QDOM_DEFAULT, binaryTypeInfo, 2, stringTypeInfo, QORE_PARAM_NO_ARG, "path", timeoutTypeInfo, QoreSimpleValue().assign(DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0)), "timeout");

    // string SFTPClient::getTextFile(string path, timeout timeout = 60s, *string encoding){}
    QC_SFTPCLIENT->addMethod("getTextFile", (q_method_n_t)SFTPClient_getTextFile_VsVtNs, Public, QCF_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 3, stringTypeInfo, QORE_PARAM_NO_ARG, "path", timeoutTypeInfo, QoreSimpleValue().assign(DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0)), "timeout", stringOrNothingTypeInfo, QORE_PARAM_NO_ARG, "encoding");

    // hash<SftpConnectionInfo> SFTPClient::info(){}
    QC_SFTPCLIENT->addMethod("info", (q_method_n_t)SFTPClient_info, Public, QCF_CONSTANT, QDOM_DEFAULT, hashdeclSftpConnectionInfo->getTypeInfo(false));

    // bool SFTPClient::isAlive(timeout timeout_ms = 60){}
    QC_SFTPCLIENT->addMethod("isAlive", (q_method_n_t)SFTPClient_isAlive_Vt, Public, QCF_NO_FLAGS, QDOM_DEFAULT, boolTypeInfo, 1, timeoutTypeInfo, QoreSimpleValue().assign((int64)60), "timeout_ms");

    // bool SFTPClient::isAliveEx(timeout timeout_ms = 60){}
    QC_SFTPCLIENT->addMethod("isAliveEx", (q_method_n_t)SFTPClient_isAliveEx_Vt, Public, QCF_NO_FLAGS, QDOM_DEFAULT, boolTypeInfo, 1, timeoutTypeInfo, QoreSimpleValue().assign((int64)60), "timeout_ms");

    // hash<SftpDirInfo> SFTPClient::list(*string path, timeout timeout = 60s){}
    QC_SFTPCLIENT->addMethod("list", (q_method_n_t)SFTPClient_list_NsVt, Public, QCF_RET_VALUE_ONLY, QDOM_DEFAULT, hashdeclSftpDirInfo->getTypeInfo(false), 2, stringOrNothingTypeInfo, QORE_PARAM_NO_ARG, "path", timeoutTypeInfo, QoreSimpleValue().assign(DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0)), "timeout");

    // list<hash<SftpFileInfo>> SFTPClient::listFull(*string path, timeout timeout = 60s){}
    QC_SFTPCLIENT->addMethod("listFull", (q_method_n_t)SFTPClient_listFull_NsVt, Public, QCF_RET_VALUE_ONLY, QDOM_DEFAULT, qore_get_complex_list_type(hashdeclSftpFileInfo->getTypeInfo(false)), 2, stringOrNothingTypeInfo, QORE_PARAM_NO_ARG, "path", timeoutTypeInfo, QoreSimpleValue().assign(DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0)), "timeout");

    // nothing SFTPClient::mkdir(string path, int mode = 0755, timeout timeout = 60s){}
    QC_SFTPCLIENT->addMethod("mkdir", (q_method_n_t)SFTPClient_mkdir_VsViVt, Public, QCF_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 3, stringTypeInfo, QORE_PARAM_NO_ARG, "path", bigIntTypeInfo, QoreSimpleValue().assign((int64)0755), "mode", timeoutTypeInfo, QoreSimpleValue().assign(DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0)), "timeout");

    // *string SFTPClient::path(){}
    QC_SFTPCLIENT->addMethod("path", (q_method_n_t)SFTPClient_path, Public, QCF_CONSTANT, QDOM_DEFAULT, stringOrNothingTypeInfo);

    // int SFTPClient::put(Qore::InputStream is, string remote_path, timeout timeout = 60s, int mode = 0644){}
    QC_SFTPCLIENT->addMethod("put", (q_method_n_t)SFTPClient_put_C11InputStreamVsVtVi, Public, QCF_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 4, QC_INPUTSTREAM->getTypeInfo(), QORE_PARAM_NO_ARG, "is", stringTypeInfo, QORE_PARAM_NO_ARG, "remote_path", timeoutTypeInfo, QoreSimpleValue().assign(DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0)), "timeout", bigIntTypeInfo, QoreSimpleValue().assign((int64)0644), "mode");

    // int SFTPClient::putFile(binary bin, string path, int mode = 0644, timeout timeout = 60s){}
    QC_SFTPCLIENT->addMethod("putFile", (q_method_n_t)SFTPClient_putFile_VyVsViVt, Public, QCF_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 4, binaryTypeInfo, QORE_PARAM_NO_ARG, "bin", stringTypeInfo, QORE_PARAM_NO_ARG, "path", bigIntTypeInfo, QoreSimpleValue().assign((int64)0644), "mode", timeoutTypeInfo, QoreSimpleValue().assign(DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0)), "timeout");

    // int SFTPClient::putFile(string data, string path, int mode = 0644, timeout timeout = 60s){}
    QC_SFTPCLIENT->addMethod("putFile", (q_method_n_t)SFTPClient_putFile_VsVsViVt, Public, QCF_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 4, stringTypeInfo, QORE_PARAM_NO_ARG, "data", stringTypeInfo, QORE_PARAM_NO_ARG, "path", bigIntTypeInfo, QoreSimpleValue().assign((int64)0644), "mode", timeoutTypeInfo, QoreSimpleValue().assign(DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0)), "timeout");

    // nothing SFTPClient::removeFile(string path, timeout timeout = 60s){}
    QC_SFTPCLIENT->addMethod("removeFile", (q_method_n_t)SFTPClient_removeFile_VsVt, Public, QCF_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, stringTypeInfo, QORE_PARAM_NO_ARG, "path", timeoutTypeInfo, QoreSimpleValue().assign(DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0)), "timeout");

    // nothing SFTPClient::rename(string old_name, string new_name, timeout timeout = 60s){}
    QC_SFTPCLIENT->addMethod("rename", (q_method_n_t)SFTPClient_rename_VsVsVt, Public, QCF_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 3, stringTypeInfo, QORE_PARAM_NO_ARG, "old_name", stringTypeInfo, QORE_PARAM_NO_ARG, "new_name", timeoutTypeInfo, QoreSimpleValue().assign(DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0)), "timeout");

    // int SFTPClient::retrieveFile(string remote_path, string local_path, timeout timeout = 60s, int mode = 0644){}
    QC_SFTPCLIENT->addMethod("retrieveFile", (q_method_n_t)SFTPClient_retrieveFile_VsVsVtVi, Public, QCF_NO_FLAGS, QDOM_FILESYSTEM, bigIntTypeInfo, 4, stringTypeInfo, QORE_PARAM_NO_ARG, "remote_path", stringTypeInfo, QORE_PARAM_NO_ARG, "local_path", timeoutTypeInfo, QoreSimpleValue().assign(DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0)), "timeout", bigIntTypeInfo, QoreSimpleValue().assign((int64)0644), "mode");

    // nothing SFTPClient::rmdir(string path, timeout timeout = 60s){}
    QC_SFTPCLIENT->addMethod("rmdir", (q_method_n_t)SFTPClient_rmdir_VsVt, Public, QCF_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, stringTypeInfo, QORE_PARAM_NO_ARG, "path", timeoutTypeInfo, QoreSimpleValue().assign(DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0)), "timeout");

    // *hash<Ssh2StatInfo> SFTPClient::stat(string path, timeout timeout = 60s){}
    QC_SFTPCLIENT->addMethod("stat", (q_method_n_t)SFTPClient_stat_VsVt, Public, QCF_RET_VALUE_ONLY, QDOM_DEFAULT, hashdeclSsh2StatInfo->getTypeInfo(true), 2, stringTypeInfo, QORE_PARAM_NO_ARG, "path", timeoutTypeInfo, QoreSimpleValue().assign(DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0)), "timeout");

    // int SFTPClient::transferFile(string local_path, string remote_path, timeout timeout = 60s, *int mode){}
    QC_SFTPCLIENT->addMethod("transferFile", (q_method_n_t)SFTPClient_transferFile_VsVsVtNi, Public, QCF_NO_FLAGS, QDOM_FILESYSTEM, bigIntTypeInfo, 4, stringTypeInfo, QORE_PARAM_NO_ARG, "local_path", stringTypeInfo, QORE_PARAM_NO_ARG, "remote_path", timeoutTypeInfo, QoreSimpleValue().assign(DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0)), "timeout", bigIntOrNothingTypeInfo, QORE_PARAM_NO_ARG, "mode");

    return QC_SFTPCLIENT;
}
