/* -*- mode: c++; indent-tabs-mode: nil -*- */
/** @file SSH2Client.qpp defines the SSH2Client class */
/*
    QC_SSH2Client.qpp

    libssh2 ssh2 client integration into qore

    Copyright 2009 Wolfgang Ritzinger
    Copyright 2010 - 2020 Qore Technologies, s.r.o.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "SSH2Client.h"

extern QoreClass* QC_SSH2BASE;
extern QoreClass* QC_SSH2CHANNEL;
TypedHashDecl* init_hashdecl_Ssh2ConnectionInfo(QoreNamespace& ns) {
    TypedHashDecl* hd = new TypedHashDecl("Ssh2ConnectionInfo", "::Qore::SSH2::Ssh2ConnectionInfo");
    hd->addMember("authenticated", stringOrNothingTypeInfo, QoreValue());
    hd->addMember("connected", boolTypeInfo, QoreValue());
    hd->addMember("fingerprint", stringOrNothingTypeInfo, QoreValue());
    hd->addMember("keyfile_priv", stringTypeInfo, QoreValue());
    hd->addMember("keyfile_pub", stringTypeInfo, QoreValue());
    hd->addMember("methods", qore_get_complex_hash_type(stringTypeInfo), QoreValue());
    hd->addMember("ssh2host", stringTypeInfo, QoreValue());
    hd->addMember("ssh2port", bigIntTypeInfo, QoreValue());
    hd->addMember("ssh2user", stringTypeInfo, QoreValue());
    ns.addSystemHashDecl(hd);
    return hd;
}

TypedHashDecl* init_hashdecl_Ssh2StatInfo(QoreNamespace& ns) {
    TypedHashDecl* hd = new TypedHashDecl("Ssh2StatInfo", "::Qore::SSH2::Ssh2StatInfo");
    hd->addMember("atime", dateTypeInfo, QoreValue());
    hd->addMember("gid", bigIntTypeInfo, QoreValue());
    hd->addMember("mode", bigIntTypeInfo, QoreValue());
    hd->addMember("mtime", dateTypeInfo, QoreValue());
    hd->addMember("permissions", stringTypeInfo, QoreValue());
    hd->addMember("size", bigIntTypeInfo, QoreValue());
    hd->addMember("uid", bigIntTypeInfo, QoreValue());
    ns.addSystemHashDecl(hd);
    return hd;
}

/* Qore class Qore::SSH2::SSH2Client */

qore_classid_t CID_SSH2CLIENT;
QoreClass* QC_SSH2CLIENT;

// SSH2Client::constructor(string url, softint port = 0) {}
static void SSH2Client_constructor_Vsvi(QoreObject* self, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* url = HARD_QORE_VALUE_STRING(args, 0);
    int64 port = HARD_QORE_VALUE_INT(args, 1);
# 105 "QC_SSH2Client.qpp"
QoreURL qurl(url);

    if (!qurl.getHost()) {
        xsink->raiseException("SSH2CLIENT-PARAMETER-ERROR", "no hostname found in URL '%s'", url->getBuffer());
        return;
    }

    if (qurl.getProtocol() && strcasecmp("ssh", qurl.getProtocol()->getBuffer()) && strcasecmp("ssh2", qurl.getProtocol()->getBuffer())) {
        xsink->raiseException("SSH2CLIENT-PARAMETER-ERROR", "URL given in the first argument to SSH2Client::constructor() specifies invalid protocol '%s' (expecting 'ssh' or 'ssh2')", qurl.getProtocol()->getBuffer());
        return;
    }

    // create private data object
    SSH2Client* mySSH2Client = new SSH2Client(qurl, (int)port);
    self->setPrivate(CID_SSH2CLIENT, mySSH2Client);
}

// SSH2Client::copy() {}
static void SSH2Client_copy(QoreObject* self, QoreObject* old, SSH2Client* c, ExceptionSink* xsink) {
# 126 "QC_SSH2Client.qpp"
xsink->raiseException("SSH2CLIENT-COPY-ERROR", "copying ssh2 connection objects is not allowed");
}

// hash<Ssh2ConnectionInfo> SSH2Client::info(){}
static QoreValue SSH2Client_info(QoreObject* self, SSH2Client* c, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 147 "QC_SSH2Client.qpp"
    return c->sshInfo(hashdeclSsh2ConnectionInfo, xsink);
}

// SSH2Channel SSH2Client::openDirectTcpipChannel(string host, softint port, string source_host = "127.0.0.1", softint source_port = 22, timeout timeout = 60s){}
static QoreValue SSH2Client_openDirectTcpipChannel_VsviVsviVt(QoreObject* self, SSH2Client* c, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* host = HARD_QORE_VALUE_STRING(args, 0);
    int64 port = HARD_QORE_VALUE_INT(args, 1);
    const QoreStringNode* source_host = HARD_QORE_VALUE_STRING(args, 2);
    int64 source_port = HARD_QORE_VALUE_INT(args, 3);
    int64 timeout = HARD_QORE_VALUE_INT(args, 4);
# 180 "QC_SSH2Client.qpp"
    static const char *SSH2CLIENT_OPENDIRECTTCPIPCHANNEL_ERR = "SSH2CLIENT-OPENDIRECTTCPIPCHANNEL-ERROR";

    if (!port) {
        xsink->raiseException(SSH2CLIENT_OPENDIRECTTCPIPCHANNEL_ERR, "port number for forwarded channel as second argument to SSH2Client::openDirectTcpipChannel() cannot be zero");
        return QoreValue();
    }

    if (!source_port) {
        xsink->raiseException(SSH2CLIENT_OPENDIRECTTCPIPCHANNEL_ERR, "source port number as fourth argument to SSH2Client::openDirectTcpipChannel() cannot be zero");
        return QoreValue();
    }

    return c->openDirectTcpipChannel(xsink, host->getBuffer(), port, source_host->getBuffer(), source_port, timeout);
}

// SSH2Channel SSH2Client::openSessionChannel(timeout timeout = 60s){}
static QoreValue SSH2Client_openSessionChannel_Vt(QoreObject* self, SSH2Client* c, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    int64 timeout = HARD_QORE_VALUE_INT(args, 0);
# 161 "QC_SSH2Client.qpp"
    return c->openSessionChannel(xsink, timeout);
}

// SSH2Channel SSH2Client::scpGet(string path, timeout timeout = 60s, *reference<hash<Ssh2StatInfo>> statinfo){}
static QoreValue SSH2Client_scpGet_VsVtC30_reference_hash_Ssh2StatInfo__(QoreObject* self, SSH2Client* c, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* path = HARD_QORE_VALUE_STRING(args, 0);
    int64 timeout = HARD_QORE_VALUE_INT(args, 1);
    const ReferenceNode* statinfo = get_param_value(args, 2).get<const ReferenceNode>();
# 217 "QC_SSH2Client.qpp"
    ReferenceHolder<QoreHashNode> ref(statinfo ? new QoreHashNode(hashdeclSsh2StatInfo, xsink) : nullptr, xsink);

    ReferenceHolder<QoreObject> o(c->scpGet(xsink, path->c_str(), timeout, *ref), xsink);
    if (o && statinfo) {
        QoreTypeSafeReferenceHelper rh(statinfo, xsink);
        // a deadlock exception occurred accessing the reference's value pointer
        if (!rh || rh.assign(ref.release()))
            return QoreValue();
    }
    return o.release();
}

// nothing SSH2Client::scpGet(string path, Qore::OutputStream os, timeout timeout = 60s){}
static QoreValue SSH2Client_scpGet_VsC12OutputStreamVt(QoreObject* self, SSH2Client* c, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* path = HARD_QORE_VALUE_STRING(args, 0);
    HARD_QORE_VALUE_OBJ_DATA(os, OutputStream, args, 1, CID_OUTPUTSTREAM, "SSH2Client::scpGet()", "Qore::OutputStream", xsink);
    if (*xsink)
        return 0;
    int64 timeout = HARD_QORE_VALUE_INT(args, 2);
# 244 "QC_SSH2Client.qpp"
    SimpleRefHolder<OutputStream> osHolder(os);
    c->scpGet(xsink, path->getBuffer(), os, timeout);
    return QoreValue();
}

// SSH2Channel SSH2Client::scpPut(string remote_path, softint size, softint mode = 0644, *date mtime, *date atime, timeout timeout = 60s){}
static QoreValue SSH2Client_scpPut_VsviviNdNdVt(QoreObject* self, SSH2Client* c, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* remote_path = HARD_QORE_VALUE_STRING(args, 0);
    int64 size = HARD_QORE_VALUE_INT(args, 1);
    int64 mode = HARD_QORE_VALUE_INT(args, 2);
    const DateTimeNode* mtime = get_param_value(args, 3).get<const DateTimeNode>();
    const DateTimeNode* atime = get_param_value(args, 4).get<const DateTimeNode>();
    int64 timeout = HARD_QORE_VALUE_INT(args, 5);
# 266 "QC_SSH2Client.qpp"
    static const char *SSH2CLIENT_SCPPUT_ERR = "SSH2CLIENT-SCPPUT-ERROR";

    if (size <= 0) {
        xsink->raiseException(SSH2CLIENT_SCPPUT_ERR, "invalid file size as second argument to SSH2Client::scpPut() (got invalid size " QLLD ")", size);
        return QoreValue();
    }

    return c->scpPut(xsink, remote_path->getBuffer(), size, (int)mode, mtime ? mtime->getEpochSeconds() : 0, atime ? atime->getEpochSeconds() : 0, timeout);
}

// nothing SSH2Client::scpPut(string remote_path, Qore::InputStream is, softint size, softint mode = 0644, *date mtime, *date atime, timeout timeout = 60s){}
static QoreValue SSH2Client_scpPut_VsC11InputStreamviviNdNdVt(QoreObject* self, SSH2Client* c, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* remote_path = HARD_QORE_VALUE_STRING(args, 0);
    HARD_QORE_VALUE_OBJ_DATA(is, InputStream, args, 1, CID_INPUTSTREAM, "SSH2Client::scpPut()", "Qore::InputStream", xsink);
    if (*xsink)
        return 0;
    int64 size = HARD_QORE_VALUE_INT(args, 2);
    int64 mode = HARD_QORE_VALUE_INT(args, 3);
    const DateTimeNode* mtime = get_param_value(args, 4).get<const DateTimeNode>();
    const DateTimeNode* atime = get_param_value(args, 5).get<const DateTimeNode>();
    int64 timeout = HARD_QORE_VALUE_INT(args, 6);
# 295 "QC_SSH2Client.qpp"
    SimpleRefHolder<InputStream> isHolder(is);
    static const char *SSH2CLIENT_SCPPUT_ERR = "SSH2CLIENT-SCPPUT-ERROR";

    if (size <= 0) {
        xsink->raiseException(SSH2CLIENT_SCPPUT_ERR, "invalid file size as third argument to SSH2Client::scpPut() (got invalid size " QLLD ")", size);
        return QoreValue();
    }

    c->scpPut(xsink, remote_path->getBuffer(), is, size, (int)mode, mtime ? mtime->getEpochSeconds() : 0, atime ? atime->getEpochSeconds() : 0, timeout);
    return QoreValue();
}

DLLLOCAL void preinitSSH2ClientClass() {
    QC_SSH2CLIENT = new QoreClass("SSH2Client", "::Qore::SSH2::SSH2Client", QDOM_NETWORK);
    CID_SSH2CLIENT = QC_SSH2CLIENT->getID();
    QC_SSH2CLIENT->setSystem();
}

DLLLOCAL QoreClass* initSSH2ClientClass(QoreNamespace& ns) {
    if (!QC_SSH2CLIENT)
        preinitSSH2ClientClass();

    // set parent class
    assert(QC_SSH2BASE);
    QC_SSH2CLIENT->addBuiltinVirtualBaseClass(QC_SSH2BASE);

    // SSH2Client::constructor(string url, softint port = 0) {}
    QC_SSH2CLIENT->addConstructor(SSH2Client_constructor_Vsvi, Public, QCF_NO_FLAGS, QDOM_DEFAULT, 2, stringTypeInfo, QORE_PARAM_NO_ARG, "url", softBigIntTypeInfo, QoreSimpleValue().assign((int64)0), "port");

    // SSH2Client::copy() {}
    QC_SSH2CLIENT->setCopy((q_copy_t)SSH2Client_copy);

    // hash<Ssh2ConnectionInfo> SSH2Client::info(){}
    QC_SSH2CLIENT->addMethod("info", (q_method_n_t)SSH2Client_info, Public, QCF_CONSTANT, QDOM_DEFAULT, hashdeclSsh2ConnectionInfo->getTypeInfo(false));

    // SSH2Channel SSH2Client::openDirectTcpipChannel(string host, softint port, string source_host = "127.0.0.1", softint source_port = 22, timeout timeout = 60s){}
    QC_SSH2CLIENT->addMethod("openDirectTcpipChannel", (q_method_n_t)SSH2Client_openDirectTcpipChannel_VsviVsviVt, Public, QCF_NO_FLAGS, QDOM_DEFAULT, QC_SSH2CHANNEL->getTypeInfo(), 5, stringTypeInfo, QORE_PARAM_NO_ARG, "host", softBigIntTypeInfo, QORE_PARAM_NO_ARG, "port", stringTypeInfo, QoreSimpleValue().assign(new QoreStringNode("127.0.0.1")), "source_host", softBigIntTypeInfo, QoreSimpleValue().assign((int64)22), "source_port", timeoutTypeInfo, QoreSimpleValue().assign(DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0)), "timeout");

    // SSH2Channel SSH2Client::openSessionChannel(timeout timeout = 60s){}
    QC_SSH2CLIENT->addMethod("openSessionChannel", (q_method_n_t)SSH2Client_openSessionChannel_Vt, Public, QCF_NO_FLAGS, QDOM_DEFAULT, QC_SSH2CHANNEL->getTypeInfo(), 1, timeoutTypeInfo, QoreSimpleValue().assign(DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0)), "timeout");

    // SSH2Channel SSH2Client::scpGet(string path, timeout timeout = 60s, *reference<hash<Ssh2StatInfo>> statinfo){}
    QC_SSH2CLIENT->addMethod("scpGet", (q_method_n_t)SSH2Client_scpGet_VsVtC30_reference_hash_Ssh2StatInfo__, Public, QCF_NO_FLAGS, QDOM_DEFAULT, QC_SSH2CHANNEL->getTypeInfo(), 3, stringTypeInfo, QORE_PARAM_NO_ARG, "path", timeoutTypeInfo, QoreSimpleValue().assign(DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0)), "timeout", qore_get_complex_reference_or_nothing_type(hashdeclSsh2StatInfo->getTypeInfo(false)), QORE_PARAM_NO_ARG, "statinfo");

    // nothing SSH2Client::scpGet(string path, Qore::OutputStream os, timeout timeout = 60s){}
    QC_SSH2CLIENT->addMethod("scpGet", (q_method_n_t)SSH2Client_scpGet_VsC12OutputStreamVt, Public, QCF_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 3, stringTypeInfo, QORE_PARAM_NO_ARG, "path", QC_OUTPUTSTREAM->getTypeInfo(), QORE_PARAM_NO_ARG, "os", timeoutTypeInfo, QoreSimpleValue().assign(DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0)), "timeout");

    // SSH2Channel SSH2Client::scpPut(string remote_path, softint size, softint mode = 0644, *date mtime, *date atime, timeout timeout = 60s){}
    QC_SSH2CLIENT->addMethod("scpPut", (q_method_n_t)SSH2Client_scpPut_VsviviNdNdVt, Public, QCF_NO_FLAGS, QDOM_DEFAULT, QC_SSH2CHANNEL->getTypeInfo(), 6, stringTypeInfo, QORE_PARAM_NO_ARG, "remote_path", softBigIntTypeInfo, QORE_PARAM_NO_ARG, "size", softBigIntTypeInfo, QoreSimpleValue().assign((int64)0644), "mode", dateOrNothingTypeInfo, QORE_PARAM_NO_ARG, "mtime", dateOrNothingTypeInfo, QORE_PARAM_NO_ARG, "atime", timeoutTypeInfo, QoreSimpleValue().assign(DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0)), "timeout");

    // nothing SSH2Client::scpPut(string remote_path, Qore::InputStream is, softint size, softint mode = 0644, *date mtime, *date atime, timeout timeout = 60s){}
    QC_SSH2CLIENT->addMethod("scpPut", (q_method_n_t)SSH2Client_scpPut_VsC11InputStreamviviNdNdVt, Public, QCF_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 7, stringTypeInfo, QORE_PARAM_NO_ARG, "remote_path", QC_INPUTSTREAM->getTypeInfo(), QORE_PARAM_NO_ARG, "is", softBigIntTypeInfo, QORE_PARAM_NO_ARG, "size", softBigIntTypeInfo, QoreSimpleValue().assign((int64)0644), "mode", dateOrNothingTypeInfo, QORE_PARAM_NO_ARG, "mtime", dateOrNothingTypeInfo, QORE_PARAM_NO_ARG, "atime", timeoutTypeInfo, QoreSimpleValue().assign(DateTimeNode::makeRelative(0, 0, 0, 0, 0, 60, 0)), "timeout");

    return QC_SSH2CLIENT;
}
